#ifndef DALI_GRAPHICS_GLES_RESOURCE_H
#define DALI_GRAPHICS_GLES_RESOURCE_H

/*
 * Copyright (c) 2021 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

namespace Dali::Graphics
{
class EglGraphicsController;

namespace GLES
{
/**
 * @brief Base class for the Graphics resource.
 */
template<class BASE, class CreateInfo>
class Resource : public BASE
{
public:
  /**
   * @brief Constructor
   * @param[in] createInfo CreateInfo structure for the resource
   * @param[in] controller Reference to the controller
   */
  Resource(const CreateInfo& createInfo, Graphics::EglGraphicsController& controller)
  : mCreateInfo(createInfo),
    mController(controller)
  {
  }

  /**
   * @brief Destructor
   */
  ~Resource() override = default;

  /**
   * @brief Retrieves create info structure
   * @return Reference to the create info structure
   */
  [[nodiscard]] const CreateInfo& GetCreateInfo() const
  {
    return mCreateInfo;
  }

  /**
   * @brief Retrieves controller
   * @return Reference to the controller object
   */
  [[nodiscard]] Graphics::EglGraphicsController& GetController() const
  {
    return mController;
  }

  /**
   * @brief Destroys resource
   *
   * This function must be implemented by the derived class.
   * It should perform final destruction of used GL resources.
   */
  virtual void DestroyResource() = 0;

  /**
   * @brief Initializes resource
   *
   * This function must be implemented by the derived class.
   * It should initialize all necessary GL resources.
   *
   * @return True on success
   */
  virtual bool InitializeResource() = 0;

  /**
   * @brief Discards resource by adding it to the discard queue
   */
  virtual void DiscardResource() = 0;

  /**
   * @brief returns pointer to base
   * @return
   */
  BASE* GetBase()
  {
    return this;
  }

protected:
  CreateInfo                       mCreateInfo; ///< CreateInfo structure
  Graphics::EglGraphicsController& mController; ///< Reference to the Controller object
};
} // namespace GLES
} // namespace Dali::Graphics
#endif
